@@ -270,7 +270,10 @@ class SampledField(Field):
             extrapolation_ = operator(self._extrapolation, other.extrapolation)
             return self.with_values(values).with_extrapolation(extrapolation_)
         else:
-            other = math.tensor(other)
+            if isinstance(other, (tuple, list)) and len(other) == self.spatial_rank:
+                other = math.tensor(other, self.points.shape['vector'])
+            else:
+                other = math.tensor(other)
             values = operator(self._values, other)
             return self.with_values(values)
 
@@ -349,6 +352,7 @@ def reduce_sample(field: Field, geometry: Geometry, dim=channel('vector')) -> ma
             sampled = [c._sample(p) for c, p in zip(components, geometry.unstack(geometry.shape.channel.name))]
         else:
             sampled = [field._sample(p) for p in geometry.unstack(geometry.shape.channel.name)]
+        dim = dim._with_item_names(geometry.shape.channel.item_names)
         return math.stack(sampled, dim)
     else:  # Nothing to reduce
         return field._sample(geometry)
