@@ -70,16 +70,17 @@ class SignatureKey:
         return shape_.with_sizes(sizes)
 
 
-def match_output_signature(new_in: SignatureKey, recorded_mappings: Dict[SignatureKey, SignatureKey]) -> SignatureKey:
+def match_output_signature(new_in: SignatureKey, recorded_mappings: Dict[SignatureKey, SignatureKey], source) -> SignatureKey:
     for rec_in, rec_out in recorded_mappings.items():
         if rec_in == new_in:  # exact match
             return rec_out
     for rec_in, rec_out in recorded_mappings.items():
         if rec_in.matches_structure_and_names(new_in):
             return rec_out.extrapolate(rec_in, new_in)
-    raise KeyError(f"Not output shape found for input shapes {new_in}. "
-                   f"Maybe the backend extrapolated the concrete function from another trace? "
-                   f"Registered transforms: {recorded_mappings}")
+    transforms_str = ''.join([f'\n* {i} -> {o}' for i, o in recorded_mappings.items()])
+    raise RuntimeError(f"{source}: no output shape found for input shapes {new_in}.\n"
+                       f"Maybe the backend extrapolated the concrete function from another trace?\n"
+                       f"Registered transforms:\n{transforms_str}")  # KeyError does not support \n
 
 
 def key_from_args(*args, cache=False, **kwargs) -> Tuple[SignatureKey, list]:
@@ -146,7 +147,7 @@ class JitFunction:
         if key not in self.traces:
             self.traces[key] = self._jit_compile(key)
         native_result = self.traces[key](*natives)
-        output_key = match_output_signature(key, self.recorded_mappings)
+        output_key = match_output_signature(key, self.recorded_mappings, self)
         output_tensors = assemble_tensors(native_result, output_key.shapes)
         return assemble_tree(output_key.tree, output_tensors)
 
@@ -276,7 +277,7 @@ class LinearFunction(Generic[X, Y], Callable[[X], Y]):
         for i, c_arg in enumerate(condition_args):
             kwargs[f'_condition_arg[{i}]'] = c_arg
         key, _ = key_from_args(x, cache=False, **kwargs)
-        assert key.backend.supports(Backend.sparse_coo_tensor)
+        # assert key.backend.supports(Backend.sparse_coo_tensor)
         return key
 
     def stencil_inspector(self, *args, **kwargs):
@@ -326,7 +327,7 @@ class GradientFunction:
         self.f = f
         self.wrt = wrt
         self.get_output = get_output
-        self.grads: Dict[SignatureKey, Callable] = {}
+        self.traces: Dict[SignatureKey, Callable] = {}
         self.recorded_mappings: Dict[SignatureKey, SignatureKey] = {}
         self.jit = jit
 
@@ -356,10 +357,10 @@ class GradientFunction:
                 raise AssertionError(f"functional_gradient() not supported by {key.backend}.")
         wrt_tensors = self._track_wrt(args)
         wrt_natives = self._track_wrt_natives(wrt_tensors, disassemble_tree(args)[1])
-        if key not in self.grads:
-            self.grads[key] = self._trace_grad(key, wrt_natives)
-        native_result = self.grads[key](*natives)
-        output_key = match_output_signature(key, self.recorded_mappings)
+        if key not in self.traces:
+            self.traces[key] = self._trace_grad(key, wrt_natives)
+        native_result = self.traces[key](*natives)
+        output_key = match_output_signature(key, self.recorded_mappings, self)
         if self.get_output:
             result_shapes = list(output_key.shapes) + [key.shapes[i] for i in wrt_tensors]
             output_tensors = assemble_tensors(native_result, result_shapes)
@@ -441,7 +442,7 @@ class HessianFunction:
         self.get_output = get_output
         self.get_gradient = get_gradient
         self.dim_suffixes = dim_suffixes
-        self.grads: Dict[SignatureKey, Callable] = {}
+        self.traces: Dict[SignatureKey, Callable] = {}
         self.recorded_mappings: Dict[SignatureKey, SignatureKey] = {}
         self.jit = jit
 
@@ -471,11 +472,11 @@ class HessianFunction:
                 raise AssertionError(f"functional_gradient() not supported by {key.backend}.")
         wrt_tensors: List[int] = self._track_wrt(args)
         wrt_natives: List[int] = self._track_wrt_natives(wrt_tensors, disassemble_tree(args)[1])
-        if key not in self.grads:
-            self.grads[key] = self._trace_hessian(key, wrt_natives)
-        native_result = self.grads[key](*natives)
+        if key not in self.traces:
+            self.traces[key] = self._trace_hessian(key, wrt_natives)
+        native_result = self.traces[key](*natives)
         assert len(native_result) == 1 + int(self.get_output) + int(self.get_gradient)
-        output_key = match_output_signature(key, self.recorded_mappings)
+        output_key = match_output_signature(key, self.recorded_mappings, self)
         result = ()
         if self.get_output:
             output_tensors = assemble_tensors(native_result[0], output_key.shapes)
@@ -625,7 +626,7 @@ class CustomGradientFunction:
             if len(self.traces) >= 8:
                 warnings.warn(f"{self.__name__} has been traced {len(self.traces)} times. To avoid memory leaks, call {self.f.__name__}.traces.clear(), {self.f.__name__}.recorded_mappings.clear()", RuntimeWarning, stacklevel=2)
         native_result = self.traces[key](*natives)
-        output_key = match_output_signature(key, self.recorded_mappings)
+        output_key = match_output_signature(key, self.recorded_mappings, self)
         output_tensors = assemble_tensors(native_result, output_key.shapes)
         return assemble_tree(output_key.tree, output_tensors)
 
@@ -812,7 +813,7 @@ class ShiftLinTracer(Tensor):
         if coo.values.nnz.size != len(scipy_csr.data):
             warnings.warn("Failed to create CSR matrix because the CSR matrix contains fewer non-zero values than COO. This can happen when the `x` tensor is too small for the stencil.", RuntimeWarning)
             return coo
-        values = coo.values.nnz[scipy_csr.data - 1]  # Change order accordingly
+        values = coo.values.nnz[wrap(scipy_csr.data - 1, instance('nnz'))]  # Change order accordingly
         self._sparse_csr = SparseMatrixContainer('csr', coo.shape, coo.indices_key, coo.src_shape, values, row_ptr, col_indices)
         return self._sparse_csr
 
@@ -832,19 +833,26 @@ class ShiftLinTracer(Tensor):
         if coo.values.nnz.size != len(scipy_csr.data):
             warnings.warn("Failed to create CSR matrix because the CSR matrix contains fewer non-zero values than COO. This can happen when the `x` tensor is too small for the stencil.", RuntimeWarning)
             return coo
-        values = coo.values.nnz[scipy_csr.data - 1]  # Change order accordingly
+        values = coo.values.nnz[wrap(scipy_csr.data - 1, instance('nnz'))]  # Change order accordingly
         self._sparse_csc = SparseMatrixContainer('csc', coo.shape, coo.indices_key, coo.src_shape, values, row_indices, col_ptr)
         return self._sparse_csc
 
     def get_sparse_matrix(self, matrix_format: str = None) -> 'SparseMatrixContainer':
-        if self.default_backend.supports(Backend.csc_matrix) and matrix_format in (None, 'csc'):
+        if matrix_format is None:
+            if self.default_backend.supports(Backend.csc_matrix):
+                matrix_format = 'csc'
+            elif self.default_backend.supports(Backend.csr_matrix):
+                matrix_format = 'csr'
+            else:
+                matrix_format = 'coo'
+        if matrix_format == 'csc':
             return self.get_sparse_csc_matrix()
-        if self.default_backend.supports(Backend.csr_matrix) and matrix_format in (None, 'csr'):
+        if matrix_format == 'csr':
             return self.get_sparse_csr_matrix()
-        elif self.default_backend.supports(Backend.sparse_coo_tensor) and matrix_format in (None, 'coo'):
+        elif matrix_format == 'coo':
             return self.get_sparse_coordinate_matrix()
         else:
-            raise NotImplementedError(f"Cannot create sparse matrix using backend '{self.default_backend}' given type '{matrix_format}'")
+            raise NotImplementedError(f"Unsupported sparse matrix format: '{matrix_format}'")
 
     @property
     def dependent_dims(self):
@@ -1063,6 +1071,7 @@ class Solve(Generic[X, Y]):  # TODO move to phi.math._functional, put Tensors th
                  x0: X or Any = None,
                  suppress: tuple or list = (),
                  preprocess_y: Callable = None,
+                 preprocess_y_args: tuple = (),
                  gradient_solve: 'Solve[Y, X]' or None = None):
         assert isinstance(method, str)
         self.method: str = method
@@ -1081,6 +1090,7 @@ class Solve(Generic[X, Y]):  # TODO move to phi.math._functional, put Tensors th
         self.preprocess_y: Callable = preprocess_y
         """ Function to be applied to the right-hand-side vector of an equation system before solving the system.
         This property is propagated to gradient solves by default. """
+        self.preprocess_y_args: tuple = preprocess_y_args
         assert all(issubclass(err, ConvergenceException) for err in suppress)
         self.suppress: tuple = tuple(suppress)
         """ Error types to suppress; `tuple` of `ConvergenceException` types. For these errors, the solve function will instead return the partial result without raising the error. """
@@ -1096,7 +1106,7 @@ class Solve(Generic[X, Y]):  # TODO move to phi.math._functional, put Tensors th
         In any case, the gradient solve information will be stored in `gradient_solve.result`.
         """
         if self._gradient_solve is None:
-            self._gradient_solve = Solve(self.method, self.relative_tolerance, self.absolute_tolerance, self.max_iterations, None, self.suppress, self.preprocess_y)
+            self._gradient_solve = Solve(self.method, self.relative_tolerance, self.absolute_tolerance, self.max_iterations, None, self.suppress, self.preprocess_y, self.preprocess_y_args)
         return self._gradient_solve
 
     def __repr__(self):
@@ -1109,12 +1119,13 @@ class Solve(Generic[X, Y]):  # TODO move to phi.math._functional, put Tensors th
                 or (self.absolute_tolerance != other.absolute_tolerance).any \
                 or (self.relative_tolerance != other.relative_tolerance).any \
                 or (self.max_iterations != other.max_iterations).any \
+                or self.preprocess_y is not other.preprocess_y \
                 or self.suppress != other.suppress:
             return False
         return self.x0 == other.x0
 
-    def __variable_attrs__(self) -> Tuple[str]:
-        return 'x0',
+    def __variable_attrs__(self):
+        return 'x0', 'preprocess_y_args'
 
 
 class SolveInfo(Generic[X, Y]):
@@ -1492,7 +1503,7 @@ def _linear_solve_forward(y, solve: Solve, native_lin_op,
                           active_dims: Shape or None, backend: Backend, is_backprop: bool) -> Any:
     PHI_LOGGER.debug(f"Performing linear solve {solve} with backend {backend}")
     if solve.preprocess_y is not None:
-        y = solve.preprocess_y(y)
+        y = solve.preprocess_y(y, *solve.preprocess_y_args)
     y_nest, (y_tensor,) = disassemble_tree(y)
     x0_nest, (x0_tensor,) = disassemble_tree(solve.x0)
     batch_dims = (y_tensor.shape & x0_tensor.shape).without(active_dims)
@@ -1573,7 +1584,7 @@ def _function_solve_forward(y, solve: Solve, f_args: tuple,
     def native_lin_f(native_x, batch_index=None):
         if batch_index is not None and batches.volume > 1:
             native_x = backend.tile(backend.expand_dims(native_x), [batches.volume, 1])
-        x = assemble_tree(x0_nest, [reshaped_tensor(native_x, [batches, active_dims] if backend.ndims(native_x) >= 2 else [active_dims])])
+        x = assemble_tree(x0_nest, [reshaped_tensor(native_x, [batches, active_dims] if backend.ndims(native_x) >= 2 else [active_dims], convert=False)])
         y = f(x, *f_args, **f_kwargs)
         _, (y_tensor,) = disassemble_tree(y)
         y_native = reshaped_native(y_tensor, [batches, active_dims] if backend.ndims(native_x) >= 2 else [active_dims])
