@@ -779,7 +779,7 @@ class Shape:
     def _with_types(self, types: 'Shape'):
         return Shape(self.sizes, self.names, tuple([types.get_type(name) if name in types else self_type for name, self_type in zip(self.names, self.types)]), self.item_names)
 
-    def _with_item_names(self, item_names):
+    def _with_item_names(self, item_names: tuple):
         return Shape(self.sizes, self.names, self.types, item_names)
 
     def _with_item_name(self, dim: str, item_name: tuple):
@@ -860,11 +860,28 @@ class Shape:
                 raise NotImplementedError(f"{type(selection)} not supported. Only (int, slice) allowed.")
         return result
 
-    def meshgrid(self):
-        """Builds a sequence containing all multi-indices within a tensor of this shape."""
+    def meshgrid(self, names=False):
+        """
+        Builds a sequence containing all multi-indices within a tensor of this shape.
+        All indices are returned as `dict` mapping dimension names to `int` indices.
+
+        The corresponding values can be retrieved from Tensors and other Sliceables using `tensor[index]`.
+
+        This function currently only supports uniform tensors.
+
+        Args:
+            names: If `True`, replace indices by their item names if available.
+
+        Returns:
+            `dict` iterator.
+        """
+        assert self.is_uniform
         indices = [0] * self.rank
         while True:
-            yield {name: index for name, index in zip(self.names, indices)}
+            if names:
+                yield {dim: (names[index] if names is not None else index) for dim, index, names in zip(self.names, indices, self.item_names)}
+            else:
+                yield {dim: index for dim, index in zip(self.names, indices)}
             for i in range(self.rank-1, -1, -1):
                 indices[i] = (indices[i] + 1) % self.sizes[i]
                 if indices[i] != 0:
@@ -984,7 +1001,7 @@ def _construct_shape(dim_type: str, *args, **dims):
     return math.Shape(sizes, names, (dim_type,) * len(sizes), item_names)
 
 
-def spatial(*args, **dims: int) -> Shape:
+def spatial(*args, **dims: int or str or tuple or list) -> Shape:
     """
     Returns the spatial dimensions of an existing `Shape` or creates a new `Shape` with only spatial dimensions.
 
@@ -1029,7 +1046,7 @@ def spatial(*args, **dims: int) -> Shape:
         raise AssertionError(f"spatial() must be called either as a selector spatial(Shape) or spatial(Tensor) or as a constructor spatial(*names, **dims). Got *args={args}, **dims={dims}")
 
 
-def channel(*args, **dims: int or str) -> Shape:
+def channel(*args, **dims: int or str or tuple or list) -> Shape:
     """
     Returns the channel dimensions of an existing `Shape` or creates a new `Shape` with only channel dimensions.
 
@@ -1078,7 +1095,7 @@ def channel(*args, **dims: int or str) -> Shape:
         raise AssertionError(f"channel() must be called either as a selector channel(Shape) or channel(Tensor) or as a constructor channel(*names, **dims). Got *args={args}, **dims={dims}")
 
 
-def batch(*args, **dims: int or str) -> Shape:
+def batch(*args, **dims: int or str or tuple or list) -> Shape:
     """
     Returns the batch dimensions of an existing `Shape` or creates a new `Shape` with only batch dimensions.
 
@@ -1123,7 +1140,7 @@ def batch(*args, **dims: int or str) -> Shape:
         raise AssertionError(f"batch() must be called either as a selector batch(Shape) or batch(Tensor) or as a constructor batch(*names, **dims). Got *args={args}, **dims={dims}")
 
 
-def instance(*args, **dims: int) -> Shape:
+def instance(*args, **dims: int or str or tuple or list) -> Shape:
     """
     Returns the instance dimensions of an existing `Shape` or creates a new `Shape` with only instance dimensions.
 
@@ -1205,7 +1222,7 @@ def merge_shapes(*shapes: Shape, order=(batch, instance, spatial, channel)):
                         raise IncompatibleShapes(f"Cannot merge shapes {shapes} because dimension '{dim.name}' exists with different sizes.", *shapes)
                     names1 = type_group.get_item_names(dim)
                     names2 = shape.get_item_names(dim)
-                    if names1 is not None and names2 is not None:
+                    if names1 is not None and names2 is not None and len(names1) > 1:
                         if names1 != names2:
                             raise IncompatibleShapes(f"Cannot merge shapes {shapes} because dimension '{dim.name}' exists with different item names.", *shapes)
                     elif names1 is None and names2 is not None:
