@@ -1,17 +1,18 @@
 import copy
 import numbers
+import traceback
 import warnings
 from typing import Tuple, Callable, List, TypeVar
 
 import numpy as np
 
-from phi.math._shape import TYPE_ABBR, IncompatibleShapes, INSTANCE_DIM, _construct_shape
-from ._config import GLOBAL_AXIS_ORDER
+from phi.math._shape import TYPE_ABBR, IncompatibleShapes, INSTANCE_DIM, _construct_shape, instance
+from ._config import GLOBAL_AXIS_ORDER, should_use_color
 from ._shape import (Shape,
                      CHANNEL_DIM, BATCH_DIM, SPATIAL_DIM, EMPTY_SHAPE,
                      parse_dim_order, shape_stack, merge_shapes, channel, concat_shapes)
 from .backend import NoBackendFound, choose_backend, BACKENDS, get_precision, default_backend, convert as convert_, \
-    Backend, PHI_LOGGER
+    Backend
 from .backend._dtype import DType
 
 
@@ -51,41 +52,76 @@ class BoundDim:
 
     @property
     def exists(self):
+        """ Whether the dimension is listed in the `Shape` of the object. """
         return self.name in self.obj.shape
 
-    def __str__(self):
-        return self.name
+    def __repr__(self):
+        if self.name not in self.obj.shape:
+            return f"{type(self.obj).__name__}.{self.name} (non-existent)"
+        items = self.item_names
+        if items is not None:
+            if len(items) <= 4:
+                size_repr = ",".join(items)
+            else:
+                size_repr = f"{self.size}:{items[0]}..{items[-1]}"
+        else:
+            size_repr = self.size
+        return f"{type(self.obj).__name__}.{self.name}{TYPE_ABBR.get(self.dim_type, '?')}={size_repr}"
 
     @property
     def size(self):
-        return self.obj.shape.get_size(self.name)
+        """ Length of this dimension as listed in the `Shape` of the bound object. """
+        return self.obj.shape.get_size(self.name) if self.exists else None
 
     @property
     def dim_type(self):
         return self.obj.shape.get_type(self.name)
 
+    @property
+    def _dim_type(self):
+        return self.obj.shape.get_type(self.name)
+
     @property
     def is_spatial(self):
-        return self.dim_type == SPATIAL_DIM
+        """ Whether the type of this dimension as listed in the `Shape` is *spatial*. Only defined for existing dimensions. """
+        return self._dim_type == SPATIAL_DIM
 
     @property
     def is_batch(self):
-        return self.dim_type == BATCH_DIM
+        """ Whether the type of this dimension as listed in the `Shape` is *batch*. Only defined for existing dimensions. """
+        return self._dim_type == BATCH_DIM
 
     @property
     def is_channel(self):
-        return self.dim_type == CHANNEL_DIM
+        """ Whether the type of this dimension as listed in the `Shape` is *channel*. Only defined for existing dimensions. """
+        return self._dim_type == CHANNEL_DIM
 
     @property
-    def is_instance(self):
-        return self.dim_type == INSTANCE_DIM
+    def item_names(self):
+        return self.obj.shape.get_item_names(self.name)
+
+    @property
+    def index(self):
+        """ The index of this dimension in the `Shape` of the `Tensor`. """
+        return self.obj.shape.index(self.name)
+
+    def __int__(self):
+        return self.index
 
     def __getitem__(self, item):
-        if isinstance(item, str):
-            item = self.obj.shape.spatial.index(item)
         return self.obj[{self.name: item}]
 
-    def unstack(self, size: int or None = None):
+    def unstack(self, size: int or None = None) -> tuple:
+        """
+        Lists the slices along this dimension as a `tuple`.
+
+        Args:
+            size: (optional) If given as `int`, this dimension can be unstacked even if it is not present on the object.
+                In that case, `size` copies of the object are returned.
+
+        Returns:
+            `tuple` of `Sliceable`
+        """
         from ._ops import unstack
         if size is None:
             return unstack(self.obj, self.name)
@@ -97,6 +133,13 @@ class BoundDim:
             else:
                 return (self.obj,) * size
 
+    def __iter__(self):
+        """ Iterate over slices along this dim """
+        if self.exists:
+            return iter(self.unstack())
+        else:
+            return iter([self.obj])
+
     def __call__(self, *args, **kwargs):
         raise TypeError(f"Method {type(self.obj).__name__}.{self.name}() does not exist.")
 
@@ -249,6 +292,12 @@ class Tensor(Sliceable):
         from ._ops import mean
         return mean(self, dim=self.shape).native()
 
+    @property
+    def std(self):
+        """ Standard deviation of this `Tensor` as a native scalar. """
+        from ._ops import std
+        return std(self, dim=self.shape).native()
+
     @property
     def sum(self):
         """ Sum of all values of this `Tensor` as a native scalar. """
@@ -292,35 +341,54 @@ class Tensor(Sliceable):
         return int(self.native())
 
     def _summary_str(self) -> str:
+        if should_use_color():
+            v = '\033[94m'  # value
+            s = '\033[92m'  # shape
+            e = '\033[0m'   # end
+            d = '\033[93m'  # dtype
+            g = '\033[37m'  # grey (additional)
+            # BOLD = '\033[1m'
+            # UNDERLINE = '\033[4m'
+        else:
+            v, s, d, e, g = '', '', '', '', ''
+
         try:
             from ._ops import all_available
             if all_available(self):
                 if self.rank == 0:
-                    return str(self.numpy())
+                    return f"{v}{str(self.numpy())}{e}"
                 elif self.shape.volume is not None and self.shape.volume <= 6:
                     content = list(np.reshape(self.numpy(self.shape.names), [-1]))
-                    content = ', '.join([repr(number) for number in content])
+                    if self.shape.rank == 1 and self.shape.get_item_names(0) is not None:
+                        content = ", ".join([f"{item}={v}{number}{e}" for number, item in zip(content, self.shape.get_item_names(0))])
+                    else:
+                        content = ', '.join([f"{v}{number}{e}" for number in content])
                     if self.shape.rank == 1 and (self.dtype.kind in (bool, int) or self.dtype.precision == get_precision()):
                         if self.shape.name == 'vector' and self.shape.type == CHANNEL_DIM:
                             return f"({content})"
-                        return f"({content}) along {self.shape.name}{TYPE_ABBR[self.shape.type]}"
-                    return f"{self.shape} {self.dtype}  {content}"
+                        return f"({content}) along {s}{self.shape.name}{TYPE_ABBR[self.shape.type]}{e}"
+                    return f"{s}{self.shape}{e} {d}{self.dtype}{e}  {content}"
                 else:
                     if self.dtype.kind in (float, int):
-                        min_val, max_val = self.min, self.max
-                        return f"{self.shape} {self.dtype}  {min_val} < ... < {max_val}"
+                        min_val, max_val, mean, std = [float(f) for f in [self.min, self.max, self.mean, self.std]]
+                        if std == 0:
+                            return f"{s}{self.shape}{e} {d}{self.dtype}{e} const {v}{mean}{e}"
+                        if any([abs(val) < 0.001 or abs(val) > 1000 for val in [mean, std]]):
+                            return f"{s}{self.shape}{e} {d}{self.dtype}{e}  {v}{mean:.2e} ± {std:.1e}{e} {g}({min_val:.0e}...{max_val:.0e}){e}"
+                        else:
+                            return f"{s}{self.shape}{e} {d}{self.dtype}{e}  {v}{mean:.3f} ± {std:.3f}{e} {g}({min_val:.0e}...{max_val:.0e}){e}"
                     elif self.dtype.kind == complex:
                         max_val = abs(self).max
-                        return f"{self.shape} {self.dtype} |...| < {max_val}"
+                        return f"{s}{self.shape}{e} {d}{self.dtype}{e} {v}|...| < {max_val}{e}"
                     elif self.dtype.kind == bool:
-                        return f"{self.shape} {self.sum} / {self.shape.volume} True"
+                        return f"{s}{self.shape}{e} {v}{self.sum} / {self.shape.volume} True{e}"
                     else:
-                        return f"{self.shape} {self.dtype}"
+                        return f"{s}{self.shape}{e} {d}{self.dtype}{e}"
             else:
                 if self.rank == 0:
-                    return f"{self.default_backend} scalar {self.dtype}"
+                    return f"{self.default_backend} scalar {d}{self.dtype}{e}"
                 else:
-                    return f"{self.default_backend} {self.shape} {self.dtype}"
+                    return f"{self.default_backend} {s}{self.shape}{e} {d}{self.dtype}{e}"
         except BaseException as err:
             return f"{self.shape}, failed to fetch values: {err}"
 
@@ -353,7 +421,44 @@ class Tensor(Sliceable):
                 warnings.warn("Slicing a Tensor with a tuple or list should only be used for channel dimensions. Use a dict or the special slicing syntax tensor.dim[slice] instead", SyntaxWarning, stacklevel=2)
                 item = {name: selection for name, selection in zip(self.shape.names, item)}
         assert isinstance(item, dict)  # dict mapping name -> slice/int
-        return self._getitem(item)
+        selections = {}
+        sliced = self
+        for dim, selection in item.items():
+            if dim not in self.shape:
+                continue
+            if isinstance(selection, Shape):
+                selection = selection.name if selection.rank == 1 else selection.names
+            if isinstance(selection, str) and ',' in selection:
+                selection = parse_dim_order(selection)
+            if isinstance(selection, str):  # single item name
+                item_names = self.shape.get_item_names(dim, fallback_spatial=True)
+                assert item_names is not None, f"No item names defined for dim '{dim}' in tensor {self.shape} and dimension size does not match spatial rank."
+                assert selection in item_names, f"Accessing tensor.{dim}['{selection}'] failed. Item names are {item_names}."
+                selection = item_names.index(selection)
+            # Either handle slicing directly or add it to the dict
+            if isinstance(selection, (tuple, list)):
+                selection_int = list(selection)
+                if any([isinstance(s, str) for s in selection]):
+                    item_names = self.shape.get_item_names(dim, fallback_spatial=True)
+                    for i, s in enumerate(selection):
+                        if isinstance(s, str):
+                            assert s in item_names, f"Accessing tensor.{dim}['{s}'] failed. Item names are {item_names}."
+                            selection_int[i] = item_names.index(s)
+                from ._ops import stack
+                result = [sliced[{dim: i}] for i in selection_int]
+                item_names = [str(n) for n in selection]
+                stack_dim = self.shape[dim] if dim in self.shape else channel(dim)
+                sliced = stack({n: r for n, r in zip(item_names, result)}, stack_dim)
+            elif isinstance(selection, Tensor) and selection.dtype.kind == bool:
+                from ._ops import boolean_mask
+                sliced = boolean_mask(sliced, dim, selection)
+            elif isinstance(selection, Tensor) and selection.dtype.kind == int:
+                from ._ops import gather
+                sliced = gather(sliced, selection, dims=dim)
+            else:
+                selections[dim] = selection
+        return sliced._getitem(selections) if selections else sliced
+
 
     def _getitem(self, selection: dict) -> 'Tensor':
         """
@@ -637,7 +742,7 @@ def shape(arg: Tensor or Shape) -> Shape:
         raise ValueError(f'shape() requires Tensor of Shape argument but got {arg}')
 
 
-class TensorDim:
+class TensorDim(BoundDim):
     """
     Reference to a specific dimension of a `Tensor`.
 
@@ -649,43 +754,8 @@ class TensorDim:
     """
 
     def __init__(self, tensor: Tensor, name: str):
+        super().__init__(tensor, name)
         self.tensor = tensor
-        self.name = name
-
-    @property
-    def exists(self):
-        """ Whether the dimension is listed in the `Shape` of the `Tensor`. """
-        return self.name in self.tensor.shape
-
-    def __str__(self):
-        """ Dimension name. """
-        return self.name
-
-    def __repr__(self):
-        return f"Dimension '{self.name}' of {self.tensor.shape}"
-
-    def unstack(self, size: int or None = None) -> tuple:
-        """
-        See `unstack_spatial()`.
-
-        Args:
-            size: (optional)
-                None: unstack along this dimension, error if dimension does not exist
-                int: repeating unstack if dimension does not exist
-
-        Returns:
-            sliced tensors
-        """
-        if size is None:
-            result = self.tensor.unstack(self.name)
-        else:
-            if self.exists:
-                unstacked = self.tensor.unstack(self.name)
-                assert len(unstacked) == size, f"Size of dimension {self.name} does not match {size}."
-                result = unstacked
-            else:
-                result = (self.tensor,) * size
-        return result
 
     def optional_unstack(self):
         """
@@ -695,6 +765,7 @@ class TensorDim:
         Returns:
             `tuple` of sliced tensors or original `Tensor`
         """
+        warnings.warn("optional_unstack() is deprecated.", DeprecationWarning)
         if self.exists:
             return self.unstack()
         else:
@@ -702,6 +773,9 @@ class TensorDim:
 
     def unstack_spatial(self, components: str or tuple or list or Shape) -> tuple:
         """
+        Deprecated.
+        Use item names instead.
+
         Slices the tensor along this dimension, returning only the selected components in the specified order.
 
         Args:
@@ -710,6 +784,7 @@ class TensorDim:
         Returns:
             selected components
         """
+        # warnings.warn(f"unstack_spatial() is deprecated. Use tensor.dim[order].dim")
         if isinstance(components, (Shape, str)):
             components = parse_dim_order(components)
         if self.exists:
@@ -724,24 +799,10 @@ class TensorDim:
             result = [self.tensor] * len(components)
         return tuple(result)
 
-    @property
-    def index(self):
-        """ The index of this dimension in the `Shape` of the `Tensor`. """
-        return self.tensor.shape.index(self.name)
-
-    def __int__(self):
-        return self.index
-
     def __len__(self):
         warnings.warn("Use Tensor.dim.size instead of len(Tensor.dim). len() only supports with integer sizes.", DeprecationWarning)
         return self.size
 
-    @property
-    def size(self):
-        """ Length of this tensor dimension as listed in the `Shape`, otherwise `1`. """
-        assert self.exists, f"Dimension '{self.name}' does not exist for tensor {self.tensor.shape}"
-        return self.tensor.shape.get_size(self.name)
-
     def as_batch(self, name: str = None):
         """ Returns a shallow copy of the `Tensor` where the type of this dimension is *batch*. """
         return self._as(BATCH_DIM, name)
@@ -780,66 +841,14 @@ class TensorDim:
         new_shape = Shape(shape.sizes, tuple(new_names), tuple(new_types), shape.item_names)
         return self.tensor._with_shape_replaced(new_shape)
 
-    @property
-    def _dim_type(self):
-        return self.tensor.shape.get_type(self.name)
-
-    @property
-    def is_spatial(self):
-        """ Whether the type of this dimension as listed in the `Shape` is *spatial*. Only defined for existing dimensions. """
-        return self._dim_type == SPATIAL_DIM
-
-    @property
-    def is_batch(self):
-        """ Whether the type of this dimension as listed in the `Shape` is *batch*. Only defined for existing dimensions. """
-        return self._dim_type == BATCH_DIM
-
-    @property
-    def is_channel(self):
-        """ Whether the type of this dimension as listed in the `Shape` is *channel*. Only defined for existing dimensions. """
-        return self._dim_type == CHANNEL_DIM
-
-    @property
-    def item_names(self):
-        return self.tensor.shape.get_item_names(self.name)
-
-    def __getitem__(self, item):
-        if isinstance(item, (int, slice, np.ndarray)):
-            return self.tensor[{self.name: item}]
-        if isinstance(item, Shape):
-            item = item.names
-        if isinstance(item, str):
-            if ',' in item:
-                item = parse_dim_order(item)
-            else:
-                if not self.exists:
-                    return self.tensor
-                item_names = self.tensor.shape.get_item_names(self.name)
-                if item_names is not None:
-                    assert item in item_names, f"Accessing tensor.{self.name}['{item}'] failed. Item names are {item_names}."
-                    item = item_names.index(item)
-                else:
-                    assert self.size == self.tensor.shape.spatial.rank, f"Cannot access tensor.{self.name}[{item}] because dimension length ({self.size}) does not match number of spatial dimensions, {self.tensor.shape.spatial}"
-                    item = self.tensor.shape.spatial.index(item)
-                return self.tensor[{self.name: item}]
-        if isinstance(item, (tuple, list)):
-            from ._ops import stack
-            result = [self[i] for i in item]
-            item_names = [str(n) for n in item]
-            stack_dim = self.tensor.shape.only(self.name) if self.exists else channel(self.name)
-            return stack({n: r for n, r in zip(item_names, result)}, stack_dim)
-        elif isinstance(item, Tensor) and item.dtype == DType(bool):
-            from ._ops import boolean_mask
-            return boolean_mask(self.tensor, self.name, item)
-        else:
-            raise ValueError(f"Syntax tensor[{type(item)}] not supported")
-
     def flip(self):
         """ Flips the element order along this dimension and returns the result as a `Tensor`. """
+        warnings.warn("dim.flip() is deprecated. Use dim[::-1] instead", DeprecationWarning)
         return self.tensor.flip(self.name)
 
     def split(self, split_dimensions: Shape):
         """ See `phi.math.unpack_dims()` """
+        warnings.warn("dim.split() is deprecated. Use math.split_dims() instead.")
         from ._ops import unpack_dims
         return unpack_dims(self.tensor, self.name, split_dimensions)
 
@@ -850,9 +859,6 @@ class TensorDim:
         else:
             return NotImplemented
 
-    def __call__(self, *args, **kwargs):
-        raise TypeError(f"Method Tensor.{self.name}() does not exist.")
-
     def sum(self):
         from ._ops import sum_
         return sum_(self.tensor, self.name)
@@ -1512,11 +1518,11 @@ def tensor(data: Tensor or Shape or tuple or list or numbers.Number,
             data = array
         else:
             inner_shape = [] if shape is None else [shape[1:]]
-            elements = [tensor(d, *inner_shape, convert=convert) for d in data]
-            common_shape = merge_shapes(*[e.shape for e in elements])
-            stack_dim = default_list_dim if shape is None else shape[0].with_sizes([len(elements)])
-            assert all(stack_dim not in t.shape for t in elements), f"Cannot stack tensors with dimension '{stack_dim}' because a tensor already has that dimension."
-            elements = [CollapsedTensor(e, common_shape) if e.shape.rank < common_shape.rank else e for e in elements]
+            tensors = [d if isinstance(d, Tensor) else tensor(d, *inner_shape, convert=convert) for d in data]
+            common_shape = merge_shapes(*[e.shape for e in tensors])
+            stack_dim = default_list_dim if shape is None else shape[0].with_sizes([len(tensors)])
+            assert all(stack_dim not in t.shape for t in tensors), f"Cannot stack tensors with dimension '{stack_dim}' because a tensor already has that dimension."
+            elements = [CollapsedTensor(e, common_shape) if e.shape.rank < common_shape.rank else e for e in tensors]
             from ._ops import cast_same
             elements = cast_same(*elements)
             return TensorStack(elements, stack_dim)
@@ -1615,7 +1621,7 @@ def compatible_tensor(data, compat_shape: Shape = None, compat_natives=(), conve
     elif isinstance(data, Shape):
         assert compat_shape.channel.rank == 1, "Only single-channel tensors support implicit casting from Shape to tensor"
         assert data.rank == compat_shape.channel.volume
-        return wrap(data.spatial.sizes, *compat_shape.channel)
+        return wrap(data.spatial.sizes, *compat_shape.channel._with_item_names((data.names,)))
     else:
         backend = choose_backend(*compat_natives, data)
         try:
