@@ -371,7 +371,7 @@ class Backend:
         """ Float tensor of selected precision containing random values in the range [0, 1) """
         raise NotImplementedError(self)
 
-    def random_normal(self, shape):
+    def random_normal(self, shape, dtype: DType):
         """ Float tensor of selected precision containing random values sampled from a normal distribution with mean 0 and std 1. """
         raise NotImplementedError(self)
 
@@ -603,7 +603,7 @@ class Backend:
         return self.cast(x, DType(int, 64))
 
     def to_complex(self, x):
-        return self.cast(x, DType(complex, max(64, min(self.precision * 2, 128))))
+        return self.cast(x, DType(complex, max(64, self.precision * 2)))
 
     def batched_gather_nd(self, values, indices):
         """
@@ -1173,6 +1173,11 @@ class Backend:
         return self.prod(self.shape(array))
 
     def multi_slice(self, tensor, slices: tuple):
+        """
+        Args:
+            tensor: value to slice
+            slices: `tuple` of `slice`, `int`, or scalar integer tensors
+        """
         return tensor[slices]
 
     def batch_gather(self, tensor, batches):
