@@ -22,7 +22,7 @@ os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'  # only errors
 if _platform.system().lower() == 'windows':  # prevent Blas GEMM launch failed on Windows
     for i, device in enumerate(_tf.config.list_physical_devices('GPU')):
         _tf.config.experimental.set_memory_growth(device, True)
-        _LOGGER.warning(f"phi.tf: Setting memory_growth on GPU {i} to True to prevent Blas errors")
+        _LOGGER.info(f"phi.tf: Setting memory_growth on GPU {i} to True to prevent Blas errors")
 
 from ._compile_cuda import compile_cuda_ops
 
