@@ -2,7 +2,7 @@ import numbers
 import warnings
 from contextlib import contextmanager
 from functools import wraps
-from typing import List, Callable
+from typing import List, Callable, Optional, Set
 
 import numpy as np
 import torch
@@ -89,6 +89,14 @@ class TorchBackend(Backend):
             tensor = tensor.to(self.get_default_device().ref)
         return tensor
 
+    def recursive_as_tensor(self, obj):
+        if isinstance(obj, (tuple, list)):
+            return type(obj)([self.recursive_as_tensor(item) for item in obj])
+        elif isinstance(obj, dict):
+            raise NotImplementedError()
+        else:
+            return self.as_tensor(obj)
+
     def auto_cast(self, *tensors) -> list:
         tensors = [t if isinstance(t, (numbers.Number, bool)) else self.as_tensor(t, True) for t in tensors]
         return Backend.auto_cast(self, *tensors)
@@ -150,86 +158,21 @@ class TorchBackend(Backend):
         return JITFunction(self, f)
 
     def custom_gradient(self, f: Callable, gradient: Callable = None) -> Callable:
-        TRACED_F = {}
-        TRACED_B = {}
+        """ See PyTorch_Jit.md """
+        def select_jit(*args):
+            if not CURRENT_JIT_CALLS:
+                return torch_function.apply(*args)
+            jit = CURRENT_JIT_CALLS[-1]
+            if torch._C._get_tracing_state() is not None:  # second call: we are tracing
+                compiled_function = jit.get_compiled_function(torch_function, args)
+                return compiled_function.apply(*args)  # this adds the compiled function to TorchScript. The function must not call any torch functions while being traced lest they be double-executed later.
+            else:  # first call: record this function
+                output = torch_function.apply(*args)
+                jit.record_autograd_function_call(torch_function, args, output)
+                return output
 
-        class TorchFunction(torch.autograd.Function):
-
-            @staticmethod
-            def forward(ctx, *args, **kwargs):
-                current_jit = CURRENT_JIT_CALLS[-1] if CURRENT_JIT_CALLS else None
-                if torch._C._get_tracing_state() is not None:
-                    assert current_jit, "Detected PyTorch tracing outside Φ-Flow"
-                    f_ = f
-
-                    def trace_later():
-                        PHI_LOGGER.debug(f"Tracing forward pass of '{f.__name__}' which uses a custom gradient")
-                        TRACED_F[current_jit] = torch.jit.trace(f, args, strict=False, check_trace=False)  # nested traces not allowed in PyTorch
-                    current_jit.post_trace.append(trace_later)
-                elif CURRENT_JIT_CALLS:
-                    jit_f = TRACED_F[CURRENT_JIT_CALLS[-1]]
-                    f_ = jit_f
-                else:
-                    f_ = f
-                y = f_(*args, **kwargs)
-                ctx.save_for_backward(*args, *y)
-                ctx.input_count = len(args)
-                ctx.jit_f = current_jit
-                return y
-
-            @staticmethod
-            def backward(ctx, *grad_args):
-                x = ctx.saved_tensors[:ctx.input_count]
-                y = ctx.saved_tensors[ctx.input_count:]
-                if torch._C._get_tracing_state() is not None:
-                    assert CURRENT_JIT_CALLS
-                    raise NotImplementedError()
-                    # g_ = gradient
-                    #
-                    # def trace_later():
-                    #     TRACED_F[current_jit] = torch.jit.trace(f, args, check_trace=False)  # nested traces not allowed in PyTorch
-                elif ctx.jit_f:
-                    jit_f = ctx.jit_f
-                    if ctx.jit_f in TRACED_B:
-                        g_ = TRACED_B[ctx.jit_f]
-                    else:
-                        # jit-compile the gradient function
-                        # jit-functions cannot return None, so we have to filter out non-required gradients
-                        needs_input_grad = ctx.needs_input_grad
-                        with ctx.jit_f:
-                            none_indices = []  # jit function cannot return None but gradient returns None to indicate there is no gradient
-
-                            def filter_required_grads(*args):
-                                grads = gradient(*args)
-                                filtered = [g for g, need in zip(grads, needs_input_grad) if need]
-                                none_indices.clear()
-                                none_indices.extend([i for i, g in enumerate(filtered) if g is None])
-                                filtered = [g for g in filtered if g is not None]
-                                assert all([isinstance(g, torch.Tensor) for g in filtered]), [type(g) for g in grads]
-                                return filtered
-
-                            needed_g = torch.jit.trace(filter_required_grads, tuple([x, y, grad_args]), check_trace=False)
-                            PHI_LOGGER.debug(f"Tracing backward pass of '{f.__name__}' which uses a custom gradient")
-
-                            def g_(*args):
-                                with jit_f:
-                                    needed = self.as_registered.call(needed_g, *args, name=f"run jit-compiled custom backward '{gradient.__name__}'")
-                                assert not jit_f.post_trace
-                                assert isinstance(needed, (tuple, list))
-                                needed = list(needed)
-                                for i in none_indices:
-                                    needed.insert(i, None)
-                                result = [(needed.pop(0) if need else None) for need in needs_input_grad]
-                                return result
-
-                            TRACED_B[ctx.jit_f] = g_
-                else:
-                    g_ = gradient
-                output = g_(x, y, grad_args)
-                result = output[0] if len(output) == 1 else (*output, )
-                return result
-
-        return TorchFunction.apply
+        torch_function = construct_torch_custom_function(f, None, None, gradient, is_f_traced=False, backend=self)
+        return select_jit
 
     def transpose(self, tensor, axes):
         return tensor.permute(axes)
@@ -251,8 +194,8 @@ class TorchBackend(Backend):
         else:
             raise ValueError(dtype)
 
-    def random_normal(self, shape):
-        return torch.randn(size=shape, dtype=to_torch_dtype(self.float_type), device=self.get_default_device().ref)
+    def random_normal(self, shape, dtype: DType):
+        return torch.randn(size=shape, dtype=to_torch_dtype(dtype or self.float_type), device=self.get_default_device().ref)
 
     def stack(self, values, axis=0):
         values = [self.as_tensor(v) for v in values]
@@ -345,6 +288,8 @@ class TorchBackend(Backend):
 
     def any(self, boolean_tensor, axis=None, keepdims=False):
         boolean_tensor = self.as_tensor(boolean_tensor, convert_external=True)
+        if self.dtype(boolean_tensor).kind != bool:
+            boolean_tensor = boolean_tensor != 0
         if axis is None:
             return torch.any(boolean_tensor)
         else:
@@ -355,6 +300,8 @@ class TorchBackend(Backend):
 
     def all(self, boolean_tensor, axis=None, keepdims=False):
         boolean_tensor = self.as_tensor(boolean_tensor, convert_external=True)
+        if self.dtype(boolean_tensor).kind != bool:
+            boolean_tensor = boolean_tensor != 0
         if axis is None:
             return torch.all(boolean_tensor)
         else:
@@ -378,6 +325,8 @@ class TorchBackend(Backend):
         return torch.where(condition, x, y)
 
     def mean(self, value, axis=None, keepdims=False):
+        if self.dtype(value).kind not in (float, complex):
+            value = self.to_float(value)
         return torch.mean(value, dim=axis, keepdim=keepdims)
 
     def range(self, start, limit=None, delta=1, dtype: DType = DType(int, 32)):
@@ -399,7 +348,11 @@ class TorchBackend(Backend):
 
     def meshgrid(self, *coordinates):
         coordinates = [self.as_tensor(c) for c in coordinates]
-        return torch.meshgrid(*coordinates, indexing='ij')
+        from packaging import version
+        if version.parse(torch.__version__) >= version.parse('1.10'):
+            return torch.meshgrid(*coordinates, indexing='ij')
+        else:
+            return torch.meshgrid(*coordinates)
 
     def linspace(self, start, stop, number):
         return torch.linspace(start, stop, number, dtype=to_torch_dtype(self.float_type), device=self.get_default_device().ref)
@@ -541,7 +494,9 @@ class TorchBackend(Backend):
         return unstacked
 
     def std(self, x, axis=None, keepdims=False):
-        return torch.std(x, dim=axis, keepdim=keepdims)
+        if self.dtype(x).kind not in (float, complex):
+            x = self.to_float(x)
+        return torch.std(x, dim=axis, keepdim=keepdims, unbiased=False)
 
     def boolean_mask(self, x, mask, axis=0):
         x = self.as_tensor(x)
@@ -697,17 +652,12 @@ class TorchBackend(Backend):
         def eval_grad(*args):
             args, wrt_args = self._prepare_graph_inputs(args, wrt)
             output = f(*args)
-            loss, aux = (output[0], output[1:]) if isinstance(output, (tuple, list)) else (output, None)
-            if loss.ndim > 0:
-                loss = loss.sum()
+            output = output if isinstance(output, (tuple, list)) else [output]
+            loss = output[0].sum()
             grads = torch.autograd.grad(loss, wrt_args)  # grad() cannot be called during jit trace
             if get_output:
-                loss = loss.detach()
-                if aux is not None:
-                    aux = [aux_.detach() if isinstance(aux_, torch.Tensor) else aux_ for aux_ in aux]
-                    return (loss, *aux, *grads)
-                else:
-                    return (loss, *grads)
+                # output[0] = output[0].detach()  # why?
+                return (*output, *grads)
             else:
                 return grads
         return eval_grad
@@ -842,55 +792,158 @@ def channels_last(x):
 
 
 class JITFunction:
+    """
+    PyTorch Tracing Procedure:
+    1. Call function non-tracing, record all called nn.Modules and autograd.Function forward calls with their args
+    2. Compile autograd.Functions forward and backward passes
+    3. Add nn.Modules to JIT Module
+    4. Trace JIT Module
+
+    Nested jit calls are ignored.
+
+    See PyTorch_Jit.md
+    """
 
     def __init__(self, backend: TorchBackend, f):
         self.backend = backend
         self.f = f
         self.traced = None
-        self.post_trace = []
+        self.autograd_function_calls = []  # (TorchCustomFunction, args, output)
+        self.compiled_functions = []  # (TorchCustomFunction, TorchCustomFunction)
+        self.autograd_function_call_counts = 0
+        self.called_modules: Set[torch.nn.Module] = set()
 
     def __call__(self, *args, **kwargs):
         if kwargs:
             raise NotImplementedError("kwargs not supported for traced function")
-        args = [self.backend.as_tensor(arg) for arg in args]
+        if CURRENT_JIT_CALLS:
+            warnings.warn(f"PyTorch does not support nested tracing. The inner JIT of {self.f.__name__} will be ignored.", RuntimeWarning)
+            return self.f(*args)
+        args = self.backend.recursive_as_tensor(args)
         if self.traced is None:
-            native_f = self.f
+            self_jit = self
+            CURRENT_JIT_CALLS.append(self)
+            self.f(*args)  # records all autograd.Function / nn.Module calls with their args -> self.autograd_function_calls, self.called_modules
+            for i, (rec_function, rec_args, rec_output) in enumerate(self.autograd_function_calls):
+                self.compiled_functions.append((rec_function, rec_function.compile(rec_args, rec_output)))
+            assert self.autograd_function_call_counts == 0
 
             class JitModule(torch.nn.Module):
 
                 def __init__(self):
                     super().__init__()
-                    for submodule in JIT_REGISTERED_MODULES:
+                    for submodule in self_jit.called_modules:
                         self.add_module(str(f"{type(submodule).__name__}_{id(submodule)}"), submodule)
 
                 def forward(self, *args):
-                    PHI_LOGGER.debug(f"Tracing Pytorch jit module for {native_f.__name__}")
-                    return native_f(*args)
+                    PHI_LOGGER.debug(f"Tracing Pytorch jit module for {self_jit.f.__name__}")
+                    return self_jit.f(*args)
 
             module = JitModule()
-            with self:
-                self.traced = torch.jit.trace(module, tuple(args), check_trace=False, strict=False)
-                # self.traced = torch.jit.trace(self.f, tuple(args), check_trace=False, strict=False)
-        with self:
-            from phi.math.backend import choose_backend
-            return choose_backend(self).call(self.traced, *args, name=f"run jit-compiled '{self.f.__name__}'")
+            self.traced = torch.jit.trace(module, tuple(args), check_trace=False, strict=False)
+            assert self.autograd_function_call_counts == len(self.autograd_function_calls), "Not all custom-gradient functions were called during tracing! Nested custom gradients are not supported."
+            assert CURRENT_JIT_CALLS.pop(-1) == self
+        from phi.math.backend import choose_backend
+        return choose_backend(self).call(self.traced, *args, name=f"run jit-compiled '{self.f.__name__}'")
+
+    def record_autograd_function_call(self, function: torch.autograd.Function, args, output):
+        self.autograd_function_calls.append((function, args, output))
+
+    def get_compiled_function(self, function: torch.autograd.Function, args) -> torch.autograd.Function:
+        assert torch._C._get_tracing_state() is not None
+        assert self.autograd_function_call_counts < len(self.autograd_function_calls), f"More custom-gradient functions were called during tracing!\nLast encountered: {function}"
+        original_function, compiled_function = self.compiled_functions[self.autograd_function_call_counts]
+        assert isinstance(compiled_function, torch.autograd.Function)
+        self.autograd_function_call_counts += 1
+        return compiled_function
 
     def __repr__(self):
-        return f"jit-TorchScript[{self.f.__name__}]"
-
-    def __enter__(self):
-        CURRENT_JIT_CALLS.append(self)
-        return self
-
-    def __exit__(self, exc_type, exc_val, exc_tb):
-        assert CURRENT_JIT_CALLS.pop(-1) == self
-        while self.post_trace:
-            self.post_trace.pop(0)()
+        return f"TorchScript[{self.f.__name__}]"
 
 
+CURRENT_JIT_CALLS: List[JITFunction] = []  # should contain no more than 1 element; PyTorch doesn't support nested tracing
 
-CURRENT_JIT_CALLS: List[JITFunction] = []
-JIT_REGISTERED_MODULES = []  # Add Networks and other modules to this list before using them in a JIT function!
+
+def register_module_call(module: torch.nn.Module):
+    if CURRENT_JIT_CALLS:
+        CURRENT_JIT_CALLS[-1].called_modules.add(module)
+
+
+def construct_torch_custom_function(f: Callable, jit_f: Optional[Callable], f_example_output, g: Callable, is_f_traced: bool, backend: TorchBackend):
+    jit_g = []
+
+    class TorchCustomFunction(torch.autograd.Function):
+        """ See PyTorch_Jit.md """
+
+        @staticmethod
+        def forward(ctx, *args, **kwargs):  # The result of this is used in the graph.
+            if torch._C._get_tracing_state():
+                PHI_LOGGER.debug(f"torch.jit.trace encountered forward pass of {f.__name__}. Returning cached output to avoid double execution.")
+                # jit_context = CURRENT_JIT_CALLS[-1]; jit_context.cached_output[torch_custom_function]
+                return f_example_output
+            y = (jit_f or f)(*args, **kwargs)
+            ctx.save_for_backward(*args, *y)
+            ctx.input_count = len(args)
+            return y
+
+        # @torch.jit.unused, @torch.jit.ignore(drop=True)  do not work here
+
+        @staticmethod
+        def backward(ctx, *grad_args):  # Breakpoints not supported here
+            x = ctx.saved_tensors[:ctx.input_count]
+            y = ctx.saved_tensors[ctx.input_count:]
+            if is_f_traced:
+                if not jit_g:
+                    # backward pass can return None but that's not allowed in JIT functions
+                    needs_input_grad = ctx.needs_input_grad
+                    none_indices = []  # jit function cannot return None but gradient returns None to indicate there is no gradient
+
+                    def filter_required_grads(*args):  # traced once
+                        grads = g(*args)
+                        filtered = [gv for gv, need in zip(grads, needs_input_grad) if need]
+                        none_indices.clear()
+                        none_indices.extend([i for i, g in enumerate(filtered) if g is None])
+                        filtered = [gv for gv in filtered if gv is not None]
+                        assert len(filtered) > 0, "Custom backward function must return at least one valid gradient."
+                        assert all([isinstance(gv, torch.Tensor) for gv in filtered]), [type(gv) for gv in grads]
+                        return filtered
+
+                    PHI_LOGGER.debug(f"Tracing backward pass of '{f.__name__}' which uses a custom gradient")
+                    needed_g = backend.jit_compile(filter_required_grads)
+                    # needed_g = torch.jit.trace(filter_required_grads, tuple([x, y, grad_args]), check_trace=False, strict=False)
+
+                    def g_(*args):  # called each time, not jitted
+                        needed = backend.as_registered.call(needed_g, *args, name=f"run jit-compiled custom backward '{g.__name__}'")
+                        assert isinstance(needed, (tuple, list))
+                        needed = list(needed)
+                        for i in none_indices:
+                            needed.insert(i, None)
+                        result = [(needed.pop(0) if need else None) for need in needs_input_grad]
+                        return result
+
+                    jit_g.append(g_)
+                output = jit_g[0](x, y, grad_args)
+            else:
+                output = g(x, y, grad_args)
+            result = output[0] if len(output) == 1 else (*output,)
+            return result
+
+        @staticmethod
+        def compile(args: tuple, output):
+            assert jit_f is None
+            PHI_LOGGER.debug(f"Tracing forward pass of '{f.__name__}' which uses a custom gradient")
+            jit_f_ = torch.jit.trace(f, args, strict=False, check_trace=False)
+            return construct_torch_custom_function(f, jit_f_, output, g, is_f_traced=True, backend=backend)
+
+        @property
+        def __name__(self):
+            return f"TorchCustomFunction-{'jit' if is_f_traced else 'non-jit'}[{f.__name__}]"
+
+        def __repr__(self):
+            return self.__name__
+
+    torch_custom_function = TorchCustomFunction()
+    return torch_custom_function
 
 
 def to_torch_dtype(dtype: DType):
