@@ -1,17 +1,18 @@
-from typing import Tuple, Any, Dict, Optional
+import warnings
+from typing import Tuple, Any, Dict, Optional, List, Callable
 
 import numpy
-import plotly.graph_objs
-from plotly import graph_objects
+import numpy as np
+from plotly import graph_objects, figure_factory
 from plotly.subplots import make_subplots
 from plotly.tools import DEFAULT_PLOTLY_COLORS
 
 from phi import math, field
 from phi.field import SampledField, PointCloud, Grid, StaggeredGrid
-from phi.geom import Sphere, BaseBox
-from phi.math import instance, Tensor, spatial
+from phi.geom import Sphere, BaseBox, Point
+from phi.math import instance, Tensor, spatial, channel
 from phi.vis._dash.colormaps import COLORMAPS
-from phi.vis._plot_util import smooth_uniform_curve
+from phi.vis._plot_util import smooth_uniform_curve, down_sample_curve
 from phi.vis._vis_base import PlottingLibrary
 
 
@@ -32,10 +33,16 @@ class PlotlyPlots(PlottingLibrary):
         fig._phi_size = size
         return fig, {pos: (pos[0]+1, pos[1]+1) for pos in subplots.keys()}
 
+    def animate(self, fig, frames: int, plot_frame_function: Callable, interval: float, repeat: bool):
+        raise NotImplementedError()
+
     def plot(self, data: SampledField, figure: graph_objects.Figure, subplot, min_val: float = None, max_val: float = None,
              show_color_bar: bool = True, **plt_args):
         _plot(data, figure, row=subplot[0], col=subplot[1], size=(800, 600), colormap=None, show_color_bar=show_color_bar)
 
+    def plotting_done(self, figure, subfigures):
+        pass
+
     def show(self, figure: graph_objects.Figure):
         figure.show()
 
@@ -58,6 +65,7 @@ def _plot(data: SampledField,
           row: int = None, col: int = None,
           ):
     subplot = fig.get_subplot(row, col)
+    vector = data.points.shape['vector']
     if data.spatial_rank == 1 and isinstance(data, Grid):
         x = data.points.vector[0].numpy().flatten()
         channels = data.values.shape.channel
@@ -67,8 +75,8 @@ def _plot(data: SampledField,
                 fig.add_trace(graph_objects.Scatter(x=x, y=y, mode='lines+markers', name=name), row=row, col=col)
             fig.update_layout(showlegend=True)
         else:
-            for channel in channels.meshgrid():
-                y = math.reshaped_native(real_values(data[channel]), [data.shape.spatial], to_numpy=True)
+            for ch_idx in channels.meshgrid():
+                y = math.reshaped_native(real_values(data[ch_idx]), [data.shape.spatial], to_numpy=True)
                 fig.add_trace(graph_objects.Scatter(x=x, y=y, mode='lines+markers', name='Multi-channel'), row=row, col=col)
             fig.update_layout(showlegend=False)
     elif data.spatial_rank == 2 and isinstance(data, Grid) and 'vector' not in data.shape:  # heatmap
@@ -86,17 +94,13 @@ def _plot(data: SampledField,
     elif data.spatial_rank == 2 and isinstance(data, Grid):  # vector field
         if isinstance(data, StaggeredGrid):
             data = data.at_centers()
-        x, y = [d.numpy('x,y') for d in data.points.vector.unstack_spatial('x,y')]
-        # ToDo Additional channel dims as multiple vectors
+        x, y = math.reshaped_native(data.points.vector[spatial(data)], [vector, data.shape.without(vector)], to_numpy=True, force_expand=True)
         extra_channels = data.shape.channel.without('vector')
-        values = math.pack_dims(real_values(data), extra_channels, math.channel('channels'))
-        data_x, data_y = [d.numpy('channels,x,y') for d in values.vector.unstack_spatial('x,y')]
+        data_x, data_y = math.reshaped_native(data.values, [vector, extra_channels, spatial(data)], to_numpy=True, force_expand=True)
         lower_x, lower_y = [float(l) for l in data.bounds.lower.vector.unstack_spatial('x,y')]
         upper_x, upper_y = [float(u) for u in data.bounds.upper.vector.unstack_spatial('x,y')]
         x_range = [lower_x, upper_x]
         y_range = [lower_y, upper_y]
-        y = y.flatten()
-        x = x.flatten()
         for ch in range(data_x.shape[0]):
             # quiver = figure_factory.create_quiver(x, y, data_x[ch], data_y[ch], scale=1.0)  # 7 points per arrow
             # fig.add_trace(quiver, row=row, col=col)
@@ -110,8 +114,8 @@ def _plot(data: SampledField,
             fig.add_scatter(x=lines_x, y=lines_y, mode='lines', row=row, col=col, name=name)
         if data_x.shape[0] == 1:
             fig.update_layout(showlegend=False)
-        fig.update_xaxes(range=x_range)
-        fig.update_yaxes(range=y_range)
+        subplot.xaxis.update(range=x_range)
+        subplot.yaxis.update(range=y_range)
         subplot.xaxis.update(scaleanchor=f'y{subplot.yaxis.plotly_name[5:]}', scaleratio=1, constrain='domain')
         subplot.yaxis.update(constrain='domain')
     elif data.spatial_rank == 3 and isinstance(data, Grid) and data.shape.channel.volume == 1:  # 3D heatmap
@@ -142,19 +146,36 @@ def _plot(data: SampledField,
                      colorscale='Blues',
                      sizemode="absolute", sizeref=1,
                      row=row, col=col)
+    elif isinstance(data, PointCloud) and data.spatial_rank == 2 and 'vector' in channel(data):
+        x, y = math.reshaped_native(data.points, [vector, data.shape.without('vector')], to_numpy=True, force_expand=True)
+        u, v = math.reshaped_native(data.values, [vector, data.shape.without('vector')], to_numpy=True, force_expand=True)
+        lower_x, lower_y = [float(d) for d in data.bounds.lower.vector]
+        upper_x, upper_y = [float(d) for d in data.bounds.upper.vector]
+        subplot.xaxis.update(range=[lower_x, upper_x])
+        subplot.yaxis.update(range=[lower_y, upper_y])
+        quiver = figure_factory.create_quiver(x, y, u, v, scale=1.0).data[0]  # 7 points per arrow
+        if data.color.shape:
+            # color = data.color.numpy(data.shape.non_channel).reshape(-1)
+            warnings.warn("Multi-colored vector plots not yet supported")
+        else:
+            color = data.color.native()
+            quiver.line.update(color=color)
+        fig.add_trace(quiver, row=row, col=col)
+        if data.points.vector.item_names:
+            subplot.xaxis.update(title=data.points.vector.item_names[0])
+            subplot.yaxis.update(title=data.points.vector.item_names[1])
+        subplot.xaxis.update(scaleanchor=f'y{subplot.yaxis.plotly_name[5:]}', scaleratio=1, constrain='domain')
+        subplot.yaxis.update(constrain='domain')
     elif isinstance(data, PointCloud) and data.spatial_rank == 2:
-        lower_x, lower_y = [float(d) for d in data.bounds.lower.vector.unstack_spatial('x,y')]
-        upper_x, upper_y = [float(d) for d in data.bounds.upper.vector.unstack_spatial('x,y')]
+        lower_x, lower_y = [float(d) for d in data.bounds.lower.vector]
+        upper_x, upper_y = [float(d) for d in data.bounds.upper.vector]
         if data.points.shape.non_channel.rank > 1:
             data_list = field.unstack(data, data.points.shape.non_channel[0].name)
             for d in data_list:
                 _plot(d, fig, size, colormap, show_color_bar, row, col)
         else:
             x, y = [d.numpy() for d in data.points.vector.unstack_spatial('x,y')]
-            if data.color.shape.instance_rank == 0:
-                color = str(data.color)
-            else:
-                color = [str(d) for d in math.unstack(data.color, instance)]
+            color = data.color.native()
             subplot_height = (subplot.yaxis.domain[1] - subplot.yaxis.domain[0]) * size[1]
             if isinstance(data.elements, Sphere):
                 symbol = 'circle'
@@ -162,14 +183,17 @@ def _plot(data: SampledField,
             elif isinstance(data.elements, BaseBox):
                 symbol = 'square'
                 marker_size = math.mean(data.elements.bounding_half_extent(), 'vector').numpy() * 1
+            elif isinstance(data.elements, Point):
+                symbol = 'x'
+                marker_size = 12 / (subplot_height / (upper_y - lower_y))
             else:
                 symbol = 'asterisk'
                 marker_size = data.elements.bounding_radius().numpy()
             marker_size *= subplot_height / (upper_y - lower_y)
             marker = graph_objects.scatter.Marker(size=marker_size, color=color, sizemode='diameter', symbol=symbol)
             fig.add_scatter(mode='markers', x=x, y=y, marker=marker, row=row, col=col)
-        fig.update_xaxes(range=[lower_x, upper_x])
-        fig.update_yaxes(range=[lower_y, upper_y])
+        subplot.xaxis.update(range=[lower_x, upper_x])
+        subplot.yaxis.update(range=[lower_y, upper_y])
         fig.update_layout(showlegend=False)
         subplot.xaxis.update(scaleanchor=f'y{subplot.yaxis.plotly_name[5:]}', scaleratio=1, constrain='domain')
         subplot.yaxis.update(constrain='domain')
@@ -182,10 +206,11 @@ def _plot(data: SampledField,
                 _plot(d, fig, size, colormap, show_color_bar, row, col)
         else:
             x, y, z = [d.numpy() for d in data.points.vector.unstack_spatial('x,y,z')]
-            if data.color.shape.instance_rank == 0:
-                color = str(data.color)
-            else:
-                color = [str(d) for d in math.unstack(data.color, instance)]
+            color = data.color.native()
+            # if data.color.shape.instance_rank == 0:
+            #     color = str(data.color)
+            # else:
+            #     color = [str(d) for d in math.unstack(data.color, instance)]
             domain_y = fig.layout[subplot.plotly_name].domain.y
             if isinstance(data.elements, Sphere):
                 symbol = 'circle'
@@ -193,6 +218,9 @@ def _plot(data: SampledField,
             elif isinstance(data.elements, BaseBox):
                 symbol = 'square'
                 marker_size = math.mean(data.elements.bounding_half_extent(), 'vector').numpy() * 1
+            elif isinstance(data.elements, Point):
+                symbol = 'x'
+                marker_size = 4 / (size[1] * (domain_y[1] - domain_y[0]) / (upper_y - lower_y) * 0.5)
             else:
                 symbol = 'asterisk'
                 marker_size = data.elements.bounding_radius().numpy()
@@ -302,14 +330,6 @@ def get_color_interpolation(val, cm_arr):
     return center
 
 
-def split_curve(x, y):
-    backtracks = numpy.argwhere(x[1:] < x[:-1])[:, 0] + 1
-    if len(backtracks) > 0:
-        x = numpy.insert(numpy.array(x, numpy.float), backtracks, numpy.nan)
-        y = numpy.insert(numpy.array(y, numpy.float), backtracks, numpy.nan)
-    return x, y
-
-
 def plot_scalars(curves: tuple or list, labels, subplots=True, log_scale='', smooth: int = 1):
     if not curves:
         return graph_objects.Figure()
@@ -331,16 +351,37 @@ def plot_scalars(curves: tuple or list, labels, subplots=True, log_scale='', smo
     return fig
 
 
-def _graph(label: str, x, y, smooth: int, index: int):
+def _graph(label: str, x: np.ndarray, y: np.ndarray, smooth: int, index: int, max_points=2000):
     color = DEFAULT_PLOTLY_COLORS[index % len(DEFAULT_PLOTLY_COLORS)]
-    x, y = split_curve(x, y)
-    if smooth > 1:
-        smooth_x, smooth_y = smooth_uniform_curve(x, y, n=smooth)
+    if len(x) > len(y):
+        x = x[:len(y)]
+    if len(y) > len(x):
+        y = y[:len(x)]
+    curves = split_curve(np.stack([x, y], -1))
+    low_res = [down_sample_curve(c, max_points) for c in curves]
+    x, y = join_curves(low_res).T
+    if smooth <= 1:
+        return [graph_objects.Scatter(x=x, y=y, name=label, line=graph_objects.scatter.Line(color=color))]
+    else:  # smooth
+        smooth_curves = [smooth_uniform_curve(c, smooth) for c in curves]
+        low_res_smooth = [down_sample_curve(c, max_points) for c in smooth_curves]
+        smooth_x, smooth_y = join_curves(low_res_smooth).T
         transparent_color = f"rgba{color[3:-1]}, 0.4)"
         return [
             graph_objects.Scatter(x=x, y=y, line=graph_objects.scatter.Line(color=transparent_color, width=1), showlegend=False),
             graph_objects.Scatter(x=smooth_x, y=smooth_y, name=label, line=graph_objects.scatter.Line(color=color, width=3), mode='lines')
         ]
-    else:
-        return [graph_objects.Scatter(x=x, y=y, name=label, line=graph_objects.scatter.Line(color=color))]
 
+
+def split_curve(curve: np.ndarray) -> List[np.ndarray]:
+    x = curve[..., 0]
+    backtracks = numpy.argwhere(x[1:] < x[:-1])[:, 0] + 1
+    if len(backtracks) == 0:
+        return [curve]
+    cuts = [0] + list(backtracks) + [curve.shape[-2]]
+    return [curve[s:e] for s, e in zip(cuts[:-1], cuts[1:])]
+
+
+def join_curves(curves: List[np.ndarray]) -> np.ndarray:
+    curves = [np.append(np.array(c, numpy.float), numpy.nan, -2) for c in curves[:-1]] + [curves[-1]]
+    return np.concatenate(curves, -2)
