@@ -1,5 +1,6 @@
 import logging
 import os
+import sys
 from numbers import Number
 from typing import Callable, Tuple, Any, Dict
 
@@ -8,14 +9,16 @@ import matplotlib.pyplot as plt
 import numpy
 import numpy as np
 from matplotlib import animation
+from matplotlib import rc
+from matplotlib.transforms import Bbox
 
 from phi import math, field
-from phi.geom import Sphere, BaseBox
-from phi.math import Tensor, batch, channel, instance, spatial
+from phi.field import Grid, StaggeredGrid, PointCloud, Scene, SampledField
+from phi.field._scene import _str
+from phi.geom import Sphere, BaseBox, Point
+from phi.math import Tensor, batch, channel, spatial, instance
 from phi.vis._plot_util import smooth_uniform_curve
 from phi.vis._vis_base import display_name, PlottingLibrary
-from phi.field import Grid, StaggeredGrid, PointCloud, Scene, unstack, SampledField
-from phi.field._scene import _str
 
 
 class MatplotlibPlots(PlottingLibrary):
@@ -45,6 +48,29 @@ class MatplotlibPlots(PlottingLibrary):
                     axes_by_pos[(row, col)] = axes[row, col]
         return figure, axes_by_pos
 
+    def animate(self, fig: plt.Figure, frames: int, plot_frame_function: Callable, interval: float, repeat: bool):
+        if 'ipykernel' in sys.modules:
+            rc('animation', html='html5')
+
+        base_axes = tuple(fig.axes)
+        positions = {a: (a.figbox.p0, a.figbox.p1) for a in base_axes}
+        specs = {a: a.get_subplotspec() for a in base_axes}
+
+        def clear_and_plot(frame: int):
+            axes = tuple(fig.axes)
+            for axis in axes:
+                if axis not in base_axes:  # colorbar etc.
+                    axis.remove()
+                else:
+                    axis.cla()  # clear
+                    box = Bbox(positions[axis])
+                    axis.set_position(box, which='active')
+                    axis.set_subplotspec(specs[axis])
+            # plt.tight_layout()
+            plot_frame_function(frame)
+
+        return animation.FuncAnimation(fig, clear_and_plot, repeat=repeat, frames=frames, interval=interval)
+
     def plot(self,
              data: SampledField,
              figure,
@@ -57,13 +83,26 @@ class MatplotlibPlots(PlottingLibrary):
         Returns:
             [Matplotlib figure](https://matplotlib.org/stable/api/figure_api.html#matplotlib.figure.Figure).
         """
-        plt.tight_layout()
+        # plt.tight_layout()
         _plot(subplot, data, show_color_bar=show_color_bar, vmin=min_val, vmax=max_val, **plt_args)
         plt.tight_layout()
         return figure
 
-    def show(self, figure: plt.Figure):
-        figure.show()
+    def plotting_done(self, figure, subfigures):
+        if isinstance(figure, plt.Figure):
+            plt.close(figure)
+        elif isinstance(figure, animation.FuncAnimation):
+            plt.close(figure._fig)
+
+    def show(self, figure):
+        if isinstance(figure, plt.Figure):
+            figure.show()
+        elif isinstance(figure, animation.FuncAnimation):
+            if 'ipykernel' in sys.modules:
+                from IPython.display import HTML
+                return HTML(figure.to_html5_video())
+            else:
+                figure._fig.show()
 
     def save(self, figure: plt.Figure, path: str, dpi: float):
         figure.savefig(path, dpi=dpi)
@@ -119,14 +158,19 @@ MATPLOTLIB = MatplotlibPlots()
 def _plot(axis, data, show_color_bar, vmin, vmax, **plt_args):
     if isinstance(data, Grid) and data.spatial_rank == 1:
         x = data.points.staggered_direction[0].vector[0].numpy()
-        for c in channel(data).meshgrid():
+        requires_legend = False
+        for c in channel(data).meshgrid(names=True):
+            label = ", ".join([i for dim, i in c.items() if isinstance(i, str)])
             values = data.values[c].numpy()
             if values.dtype in (np.complex64, np.complex128):
-                axis.plot(x, values.real, label='real')
-                axis.plot(x, values.imag, label='imag')
-                axis.legend()
+                axis.plot(x, values.real, label=f"real({label})" if label else "real")
+                axis.plot(x, values.imag, label=f"imag({label})" if label else "real")
+                requires_legend = True
             else:
-                axis.plot(x, values)
+                axis.plot(x, values, label=label)
+                requires_legend = requires_legend or label
+        if requires_legend:
+            axis.legend()
     elif isinstance(data, Grid) and channel(data).volume == 1 and data.spatial_rank == 2:
         dims = spatial(data)
         if data.bounds.upper.vector.item_names is not None:
@@ -139,7 +183,7 @@ def _plot(axis, data, show_color_bar, vmin, vmax, **plt_args):
         extent = (float(left), float(right), float(bottom), float(top))
         im = axis.imshow(data.values.numpy(dims.reversed), origin='lower', extent=extent, vmin=vmin, vmax=vmax, **plt_args)
         if show_color_bar:
-            plt.colorbar(im, ax=axis)
+            axis.figure.colorbar(im, ax=axis)  # adds a new Axis to the figure
         axis.set_xlabel(dims.names[0])
         axis.set_ylabel(dims.names[1])
     elif isinstance(data, Grid) and data.spatial_rank == 2:  # vector field
@@ -164,13 +208,30 @@ def _plot(axis, data, show_color_bar, vmin, vmax, **plt_args):
         norm = matplotlib.colors.Normalize(vmin=np.min(values), vmax=np.max(values))
         colors = cmap(norm(values))
         axis.voxels(values, facecolors=colors, edgecolor='k')
-    elif isinstance(data, PointCloud) and data.spatial_rank == 2:
+    elif isinstance(data, PointCloud) and data.spatial_rank == 2 and 'vector' in channel(data):
         axis.set_aspect('equal', adjustable='box')
-        lower_x, lower_y = [float(d) for d in data.bounds.lower.vector.unstack_spatial('x,y')]
-        upper_x, upper_y = [float(d) for d in data.bounds.upper.vector.unstack_spatial('x,y')]
+        vector = data.points.shape['vector']
+        x, y = math.reshaped_native(data.points, [vector, data.shape.without('vector')], to_numpy=True, force_expand=True)
+        u, v = math.reshaped_native(data.values, [vector, data.shape.without('vector')], to_numpy=True, force_expand=True)
+        lower_x, lower_y = [float(d) for d in data.bounds.lower.vector]
+        upper_x, upper_y = [float(d) for d in data.bounds.upper.vector]
         axis.set_xlim((lower_x, upper_x))
         axis.set_ylim((lower_y, upper_y))
-        if data.points.shape.non_channel.rank > 1:
+        if data.color.shape:
+            color = data.color.numpy(data.shape.non_channel).reshape(-1)
+        else:
+            color = data.color.native()
+        axis.quiver(x, y, u, v, color=color, units='xy', scale=1)
+        if data.points.vector.item_names:
+            axis.set_xlabel(data.points.vector.item_names[0])
+            axis.set_ylabel(data.points.vector.item_names[1])
+    elif isinstance(data, PointCloud) and data.spatial_rank == 2:
+        axis.set_aspect('equal', adjustable='box')
+        lower_x, lower_y = [float(d) for d in data.bounds.lower.vector]
+        upper_x, upper_y = [float(d) for d in data.bounds.upper.vector]
+        axis.set_xlim((lower_x, upper_x))
+        axis.set_ylim((lower_y, upper_y))
+        if data.points.shape.non_channel.rank > 1:  # multiple instance / spatial dimensions
             data_list = field.unstack(data, data.points.shape.non_channel[0].name)
             for d in data_list:
                 _plot_points(axis, d, **plt_args)
@@ -183,18 +244,21 @@ def _plot(axis, data, show_color_bar, vmin, vmax, **plt_args):
                 _plot(axis, d, show_color_bar, vmin, vmax, **plt_args)
         else:
             x, y, z = [d.numpy() for d in data.points.vector.unstack_spatial('x,y,z')]
-            color = [str(d) for d in data.color.points.unstack(len(x))]
+            color = [d.native() for d in data.color.points.unstack(len(x))]
+            M = axis.transData.get_matrix()
+            x_scale, y_scale, z_scale = M[0, 0], M[1, 1], M[2, 2]
             if isinstance(data.elements, Sphere):
                 symbol = 'o'
                 size = data.elements.bounding_radius().numpy() * 0.4
             elif isinstance(data.elements, BaseBox):
                 symbol = 's'
                 size = math.mean(data.elements.bounding_half_extent(), 'vector').numpy() * 0.35
+            elif isinstance(data.elements, Point):
+                symbol = 'x'
+                size = 6 / (0.5 * (x_scale+y_scale+z_scale)/3)
             else:
                 symbol = 'X'
                 size = data.elements.bounding_radius().numpy()
-            M = axis.transData.get_matrix()
-            x_scale, y_scale, z_scale = M[0, 0], M[1, 1], M[2, 2]
             axis.scatter(x, y, z, marker=symbol, color=color, s=(size * 0.5 * (x_scale+y_scale+z_scale)/3) ** 2)
         lower_x, lower_y, lower_z = [float(d) for d in data.bounds.lower.vector.unstack_spatial('x,y,z')]
         upper_x, upper_y, upper_z = [float(d) for d in data.bounds.upper.vector.unstack_spatial('x,y,z')]
@@ -205,20 +269,35 @@ def _plot(axis, data, show_color_bar, vmin, vmax, **plt_args):
         raise NotImplementedError(f"No figure recipe for {data}")
 
 
-def _plot_points(axis, data, **plt_args):
+def _plot_points(axis, data: PointCloud, **plt_args):
     x, y = [d.numpy() for d in data.points.vector.unstack_spatial('x,y')]
-    color = [str(d) for d in data.color.points.unstack(len(x))]
+    color = [d.native() for d in data.color.points.unstack(len(x))]
     if isinstance(data.elements, Sphere):
         symbol = 'o'
         size = data.elements.bounding_radius().numpy() * 1.41
     elif isinstance(data.elements, BaseBox):
         symbol = 's'
         size = math.mean(data.elements.bounding_half_extent(), 'vector').numpy()
+    elif isinstance(data.elements, Point):
+        symbol = 'x'
+        size = 6 / _get_pixels_per_unit(axis.figure, axis)
     else:
-        symbol = 'X'
+        symbol = '*'
         size = data.elements.bounding_radius().numpy()
     size_px = size * _get_pixels_per_unit(axis.figure, axis)
     axis.scatter(x, y, marker=symbol, color=color, s=size_px ** 2, alpha=0.8)
+    _annotate_points(axis, data.points, instance(data))
+
+
+def _annotate_points(axis, points: math.Tensor, labelled_dim: math.Shape):
+    if points.shape['vector'].size == 2:
+        x, y = math.reshaped_native(points, ['vector', points.shape.without('vector')], to_numpy=True, force_expand=True)
+        if labelled_dim.item_names[0]:
+            x_view = axis.get_xlim()[1] - axis.get_xlim()[0]
+            y_view = axis.get_ylim()[1] - axis.get_ylim()[0]
+            for x_, y_, label in zip(x, y, labelled_dim.item_names[0]):
+                axis.annotate(label, (x_ + .01 * x_view, y_ + .01 * y_view))
+
 
 
 def _get_pixels_per_unit(fig: plt.Figure, axis: plt.Axes, dpi=90):
@@ -319,7 +398,7 @@ def plot_scalars(scene: str or tuple or list or Scene or math.Tensor,
             elif names_equal:
                 axis.set_title(display_name(str(names[b])))
             elif paths_equal:
-                axis.set_title(os.path.basename(str(scene.paths[b])))
+                axis.set_title(os.path.basename(scene.paths[b].native()))
         if labels is not None:
             curve_labels = labels
         elif names_equal:
@@ -357,7 +436,8 @@ def plot_scalars(scene: str or tuple or list or Scene or math.Tensor,
                 color = cycle[int(color)]
             logging.debug(f"Plotting curve {label}")
             axis.plot(x_values, values, color=color, alpha=smooth_alpha, linewidth=1)
-            axis.plot(*smooth_uniform_curve(x_values, values, n=smooth), color=color, linewidth=smooth_linewidth, label=label)
+            curve = np.stack([x_values, values], -1)
+            axis.plot(*smooth_uniform_curve(curve, smooth), color=color, linewidth=smooth_linewidth, label=label)
             if grid:
                 if isinstance(grid, dict):
                     axis.grid(**grid)
